# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio entry point module configuration."""

from __future__ import absolute_import, print_function

import warnings

import pkg_resources


class InvenioConfigEntryPointModule(object):
    """Load configuration from module defined by entry point.

    .. versionadded:: 1.0.0
    """

    def __init__(self, app=None, entry_point_group='invenio_config.module'):
        """Initialize extension."""
        self.entry_point_group = entry_point_group
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Initialize Flask application."""
        if self.entry_point_group:
            for ep in pkg_resources.iter_entry_points(self.entry_point_group):
                app.config.from_object(ep.load())



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Default configuration loader usable by e.g. Invenio-Base."""

from __future__ import absolute_import, print_function

from .default import InvenioConfigDefault
from .entrypoint import InvenioConfigEntryPointModule
from .env import InvenioConfigEnvironment
from .folder import InvenioConfigInstanceFolder
from .module import InvenioConfigModule


def create_config_loader(config=None, env_prefix='APP'):
    """Create a default configuration loader.

    A configuration loader takes a Flask application and keyword arguments and
    updates the Flask application's configuration as it sees fit.

    This default configuration loader will load configuration in the following
    order:

        1. Load configuration from ``invenio_config.module`` entry point group.
        2. Load configuration from ``config`` module if provided as argument.
        3. Load configuration from the instance folder:
           ``<app.instance_path>/<app.name>.cfg``.
        4. Load configuration keyword arguments provided.
        5. Load configuration from environment variables with the prefix
           ``env_prefix``.

    If no secret key has been set a warning will be issued.

    :param config: Either an import string to a module with configuration or
        alternatively the module itself.
    :param env_prefix: Environment variable prefix to import configuration
        from.
    :return: A callable with the method signature
        ``config_loader(app, **kwargs)``.

    .. versionadded:: 1.0.0
    """
    def _config_loader(app, **kwargs_config):
        InvenioConfigEntryPointModule(app=app)
        if config:
            InvenioConfigModule(app=app, module=config)
        InvenioConfigInstanceFolder(app=app)
        app.config.update(**kwargs_config)
        InvenioConfigEnvironment(app=app, prefix='{0}_'.format(env_prefix))
        InvenioConfigDefault(app=app)

    return _config_loader


def create_conf_loader(*args, **kwargs):  # pragma: no cover
    """Create a default configuration loader.

    .. deprecated:: 1.0.0b1
       Use :func:`create_config_loader` instead. This function will be removed
       in version 1.0.1.
    """
    import warnings
    warnings.warn(
        '"create_conf_loader" has been renamed to "create_config_loader".',
        DeprecationWarning
    )
    return create_config_loader(*args, **kwargs)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio default configuration."""

from __future__ import absolute_import, print_function

import warnings


class InvenioConfigDefault(object):
    """Load configuration from module.

    .. versionadded:: 1.0.0
    """

    def __init__(self, app=None):
        """Initialize extension."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Initialize Flask application."""
        # Ensure SECRET_KEY is set.
        SECRET_KEY = app.config.get('SECRET_KEY')

        if SECRET_KEY is None:
            app.config['SECRET_KEY'] = 'CHANGE_ME'
            warnings.warn(
                'Set configuration variable SECRET_KEY with random string',
                UserWarning)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Simple tests."""

from __future__ import absolute_import, print_function

import os
import shutil
import tempfile
import warnings
from os.path import join

from flask import Flask
from mock import patch
from pkg_resources import EntryPoint

from invenio_config import InvenioConfigDefault, \
    InvenioConfigEntryPointModule, InvenioConfigEnvironment, \
    InvenioConfigInstanceFolder, InvenioConfigModule, create_config_loader


class ConfigEP(EntryPoint):
    """Mocking of entrypoint."""

    def __init__(self, **kwargs):
        """Save keyword arguments as config."""
        self.kwargs = kwargs

    def load(self):
        """Mock load entry point."""
        class Config(object):
            pass
        for key, val in self.kwargs.items():
            setattr(Config, key, val)
        return Config


def _mock_ep(eps):
    """Mock for pkg_resources.iter_entry_points."""
    def iter_entry_points(name):
        for ep in eps:
            yield ep
    return iter_entry_points


def test_version():
    """Test version import."""
    from invenio_config import __version__
    assert __version__


def test_module():
    """Test module."""
    app = Flask('testapp')

    class Config(object):
        TESTVAR = True

    assert not app.config.get('TESTVAR', False)
    InvenioConfigModule(app, module=Config)
    assert app.config.get('TESTVAR', False)


@patch('pkg_resources.iter_entry_points', _mock_ep([ConfigEP(TESTVAR=True)]))
def test_entry_point():
    """Test entry point."""
    app = Flask('testapp')

    assert not app.config.get('TESTVAR', False)
    InvenioConfigEntryPointModule(app)
    assert app.config.get('TESTVAR', False)


def test_folder():
    """Test instance folder loading."""
    tmppath = tempfile.mkdtemp()
    try:
        # Write config into instance folder.
        with open(join(tmppath, 'testapp.cfg'), 'w') as f:
            f.write("TESTVAR = True\n")

        app = Flask(
            'testapp', instance_path=tmppath, instance_relative_config=True)
        assert not app.config.get('TESTVAR', False)
        InvenioConfigInstanceFolder(app=app)
        assert app.config.get('TESTVAR', False)
    finally:
        shutil.rmtree(tmppath)


def test_default():
    """Test instance folder loading."""
    app = Flask('testapp')

    with warnings.catch_warnings(record=True) as w:
        assert len(w) == 0
        InvenioConfigDefault(app)
        assert len(w) == 1
        assert app.config['SECRET_KEY'] == "CHANGE_ME"

    app = Flask('testapp')
    app.config['SECRET_KEY'] = "thisisasecret"

    with warnings.catch_warnings(record=True) as w:
        assert len(w) == 0
        InvenioConfigDefault(app)
        assert len(w) == 0


def test_env():
    """Test loading from environment variables."""
    app = Flask('testapp')
    InvenioConfigEnvironment(app, prefix="MYPREFIX_")
    assert not app.config.get('TESTVAR', False)

    os.environ["MYPREFIX_TESTVAR"] = "True"
    os.environ["MYPREFIX_JUSTASTRING"] = "This is just a string"
    os.environ["MYPREFIX_COMPLEX_DICT"] = (
        "{'complex': {'python': 'dict'}, 'with': ['list', 'and', 1234]}")
    InvenioConfigEnvironment(app, prefix="MYPREFIX_")
    assert app.config.get('TESTVAR') is True
    assert app.config.get('JUSTASTRING') == "This is just a string"
    assert app.config.get('COMPLEX_DICT') == \
        {'complex': {'python': 'dict'}, 'with': ['list', 'and', 1234]}


@patch('pkg_resources.iter_entry_points', _mock_ep([
    ConfigEP(EP='ep', MODULE='ep', FOLDER='ep', KWARGS='ep', ENV='ep')]))
def test_conf_loader_factory():
    """Test the conf factory."""
    tmppath = tempfile.mkdtemp()
    try:
        app = Flask('testapp')

        # Module configuration
        class Config(object):
            MODULE = 'module'
            FOLDER = 'module'
            KWARGS = 'module'
            ENV = 'module'

        # Instance path configuration
        with open(join(tmppath, 'testapp.cfg'), 'w') as f:
            f.write("FOLDER = 'folder'\n")
            f.write("KWARGS = 'folder'\n")
            f.write("ENV = 'folder'\n")

        # Keyword arguments configuration
        kwargs = dict(KWARGS='kwargs', ENV='kwargs')

        # Environment configuraiton
        os.environ['APREFIX_ENV'] = 'env'

        # Create conf loader
        conf_loader = create_config_loader(Config, env_prefix="APREFIX")
        app = Flask(
            'testapp', instance_path=tmppath, instance_relative_config=True)
        conf_loader(app, **kwargs)

        # Test correct overwriting of values.
        assert app.config['EP'] == 'ep'
        assert app.config['MODULE'] == 'module'
        assert app.config['FOLDER'] == 'folder'
        assert app.config['KWARGS'] == 'kwargs'
        assert app.config['ENV'] == 'env'
    finally:
        shutil.rmtree(tmppath)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio configuration loader.

Invenio-Config is a *base package* of the Invenio digital library framework.
It is usually installed automatically as a dependency. It should facilitate
configuration loading from various sources to an application instance.

The following configuration loaders exists:

- :py:data:`invenio_config.default.InvenioConfigDefault` - ensure required
  configuration values are set.
- :py:data:`invenio_config.module.InvenioConfigModule` - for loading
  configuration from a Python module.
- :py:data:`invenio_config.entrypoint.InvenioConfigEntryPointModule` - for
  loading configuration from a Python module specified by an entry point (by
  default ``invenio_config.module``).
- :py:data:`invenio_config.folder.InvenioConfigInstanceFolder` - for loading
  configuration from ``cfg`` file in an instance folder.
- :py:data:`invenio_config.env.InvenioConfigEnvironment` - for loading
  configuration from environment variables with defined prefix (e.g.
  ``INVENIO_SECRET_KEY``).

It also includes configuration loader factory that it is used to merge these
sources in predefined order ensuring correct behavior in common scenarios.

Initialization
--------------
Following example needs a writable instance folder, hence we start by creating
a temporary directory.

>>> import tempfile
>>> tmppath = tempfile.mkdtemp()

.. testcode::
   :hide:

   import atexit
   import shutil
   atexit.register(lambda: shutil.rmtree(tmppath))

Now we can create a Flask application:

>>> from flask import Flask
>>> app = Flask('myapp', instance_path=tmppath, instance_relative_config=True)

Loaders
-------
You can check default configuration values in newly created ``app``.

>>> 'DEBUG' in app.config
True
>>> app.config.get('SECRET_KEY') is None
True

Default
~~~~~~~
The default configuration loader makes sure that the required configuration
values are always loaded. You should call it **after** all configuration
loaders have been already called.

For instance, the default configuration loader will warn if the ``SECRET_KEY``
is not defined:

>>> import warnings
>>> from invenio_config import InvenioConfigDefault
>>> with warnings.catch_warnings(record=True) as w:
...     config_default = InvenioConfigDefault(app=app)
...     assert len(w) == 1
>>> app.config['SECRET_KEY']
'CHANGE_ME'

Module
~~~~~~
The module loader accepts an object and proxies the call to
:meth:`flask.Config.from_object`.

Here is an example of a configuration object:

>>> class Config:
...     EXAMPLE = 'module'
>>> from invenio_config import InvenioConfigModule
>>> config_module = InvenioConfigModule(app=app, module=Config)
>>> app.config['EXAMPLE']
'module'

Entry point
~~~~~~~~~~~
The entry point loader works similar to the module loader, it just loads the
config module from the entry point ``invenio_config.module``:

>>> from invenio_config import InvenioConfigEntryPointModule
>>> config_ep = InvenioConfigEntryPointModule(app=app)

Instance Folder
~~~~~~~~~~~~~~~
The runtime configuration should be stored in a separate file, ideally located
outiside the actual application package. The configuration files are handled
as Python files where only variables in uppercase are stored in the application
config.

>>> import os
>>> from invenio_config import InvenioConfigInstanceFolder
>>> with open(os.path.join(tmppath, 'myapp.cfg'), 'w') as f:
...     result = f.write("EXAMPLE = 'instance folder'")
>>> config_instance_folder = InvenioConfigInstanceFolder(app=app)
>>> app.config['EXAMPLE']
'instance folder'

Environment
~~~~~~~~~~~
Using environment variables is very handy when it comes to configuring
connections to services like database, Redis server, RabbitMQ, etc. used via
containers (e.g. Docker). In order to protect your application from reading
environment variables set by the system or other applications, you should
define a variable prefix used by the loader.

>>> os.environ['MYAPP_EXAMPLE'] = 'environment'
>>> from invenio_config import InvenioConfigEnvironment
>>> config_environment = InvenioConfigEnvironment(app=app, prefix='MYAPP_')
>>> app.config['EXAMPLE']
'environment'

You can also set more complex Python literal variables (e.g. dictionaries or
lists):

>>> os.environ['MYAPP_COMPLEX'] = "{'items': [{'num': 42}, {'foo': 'bar'}]}"
>>> # ...or export MYAPP_COMPLEX="{'items': [{'num': 42}, {'foo': 'bar'}]}"
>>> config_environment = InvenioConfigEnvironment(app=app, prefix='MYAPP_')
>>> app.config['COMPLEX']
{'items': [{'num': 42}, {'foo': 'bar'}]}


Factory Pattern
---------------
The Invenio-Config comes with an opinionated way of loading configuration,
that combines loaders in predictable way. You can use
:func:`invenio_config.utils.create_config_loader` if you would like to:

  1. Load configuration from ``invenio_config.module`` entry point group.
  2. Load configuration from ``config`` module if provided as argument.
  3. Load configuration from the instance folder:
     ``<app.instance_path>/<app.name>.cfg``.
  4. Load configuration keyword arguments provided.
  5. Load configuration from environment variables with the prefix
     ``env_prefix``.

>>> from invenio_config import create_config_loader
>>> app = Flask('myapp', instance_path=tmppath, instance_relative_config=True)
>>> config_loader = create_config_loader(config=Config, env_prefix='MYAPP')
>>> config_loader(app=app, MYARG='config loader')
>>> app.config['EXAMPLE']
'environment'
>>> app.config['MYARG']
'config loader'

"""

from __future__ import absolute_import, print_function

from .default import InvenioConfigDefault
from .env import InvenioConfigEnvironment
from .folder import InvenioConfigInstanceFolder
from .module import InvenioConfigModule
from .entrypoint import InvenioConfigEntryPointModule
from .utils import create_conf_loader, create_config_loader
from .version import __version__

__all__ = (
    '__version__',
    'InvenioConfigDefault',
    'InvenioConfigEntryPointModule',
    'InvenioConfigEnvironment',
    'InvenioConfigInstanceFolder',
    'InvenioConfigModule',
    'create_conf_loader',
    'create_config_loader',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio instance folder configuration."""

from __future__ import absolute_import, print_function


class InvenioConfigInstanceFolder(object):
    """Load configuration from py file in folder.

    If the application have instance relative config then the file will be read
    from the instance folder, otherwise it will be read from the application
    root path.

    More about `instance folders
    <http://flask.pocoo.org/docs/latest/config/#instance-folders>`_.

    .. versionadded:: 1.0.0
    """

    def __init__(self, app=None):
        """Initialize extension."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Initialize Flask application."""
        app.config.from_pyfile('{0}.cfg'.format(app.name), silent=True)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio environment configuration."""

from __future__ import absolute_import, print_function

import ast
import os


class InvenioConfigEnvironment(object):
    """Load configuration from environment variables.

    .. versionadded:: 1.0.0
    """

    def __init__(self, app=None, prefix='INVENIO_'):
        """Initialize extension."""
        self.prefix = prefix
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Initialize Flask application."""
        prefix_len = len(self.prefix)
        for varname, value in os.environ.items():
            if not varname.startswith(self.prefix):
                continue

            # Prepare values
            varname = varname[prefix_len:]
            value = value or app.config.get(varname)

            # Evaluate value
            try:
                value = ast.literal_eval(value)
            except (SyntaxError, ValueError):
                pass

            # Set value
            app.config[varname] = value
